# -*- coding: utf-8 -*-
# author: Ethosa
from logging import getLogger, StreamHandler, Formatter

from requests import Session

from .LongPoll import LongPoll
from .VkAuthManager import VkAuthManager
from .Uploader import Uploader
from .VkScript import VkScript

from ..StartThread import StartThread


class Vk(object):
    def __init__(self, token="", group_id="",
                 login="", password="", api="5.103",
                 debug=False):
        """auth in VK

        Keyword Arguments:
            token {str} -- access_token (default: {""})
            group_id {str} -- group id if you want to log in through the group (default: {""})
            login {str} -- login. used for authorization through the user (default: {""})
            password {str} -- password. used for authorization through the user (default: {""})
            api {str} -- api version (default: {"5.103"})
            debug {bool} -- debug log (default: {False})
        """
        self.session = Session()

        # Parses vk.com, if login and password are not empty.
        if login and password:
            self.auth = VkAuthManager()
            self.auth.login(login, password)
            token = self.auth.get_token()

        self.v = api
        self.token = token
        self.group_id = group_id

        self.method = ""
        self.events = {}

        # Debug settings.
        self.debug = 50
        if debug:
            if isinstance(debug, int):
                self.debug = debug
            else:
                self.debug = 10

        # Logger initialize.
        self.logger = getLogger("saya")
        self.logger.setLevel(self.debug)
        handler = StreamHandler()
        handler.setFormatter(
            Formatter("[%(levelname)s] %(name)s: %(asctime)s — %(message)s")
        )
        self.logger.addHandler(handler)

        self.uploader = Uploader(self)
        self.longpoll = LongPoll(self)
        self.vks = VkScript()  # for pyexecute method.

    def call_method(self, method, data={}):
        """call to any method in VK api

        Arguments:
            method {str} -- method name
            e.g. "messages.send", "wall.post"

        Keyword Arguments:
            data {dict} -- data to send (default: {{}})

        Returns:
            dict -- response after calling method
        """
        data["v"] = self.v
        data["access_token"] = self.token
        response = self.session.post(
                "https://api.vk.com/method/%s" % method, data=data
            ).json()

        # Logging.
        if "error" in response:
            self.logger.error('Error [%s] in called method "%s": %s' % (
                    response["error"]["error_code"], method,
                    response["error"]["error_msg"]
                )
            )
        else:
            self.logger.debug('Successfully called method "%s"' % (method))
        return response

    def execute(self, code):
        """
        Calls an execute VK API method

        Arguments:
            code {str} -- VKScript code.

        Returns:
            dict -- response
        """
        return self.call_method("execute", {"code": code})

    def pyexecute(self, code):
        """
        Calls an execute VK API method

        Arguments:
            code {str} -- Python code.

        Returns:
            dict -- response
        """
        return self.execute(self.vks.translate(code))

    def start_listen(self):
        """
        Starts receiving events from the server.
        """
        for event in self.longpoll.listen(True):
            if "type" in event:
                if event["type"] in self.events:
                    self.events[event["type"]](event)
                elif event["type"] in dir(self):
                    getattr(self, event["type"])(event)
            else:
                self.logger.warning('Unknown event passed: "%s"' % (event))

    def __getattr__(self, attr):
        """A convenient alternative for the call_method method.

        Arguments:
            attr {str} -- method name
            e.g. messages.send, wall.post

        Returns:
            response after calling method
        """
        if attr.startswith("on_"):  # e.g. on_message_new
            attr = attr[3:]

            def decorator(obj):
                def listen(f):
                    for event in self.longpoll.listen(True):
                        if event["type"] == attr:
                            obj(event)

                if callable(obj):
                    StartThread(listen, obj).start()
                else:
                    # obj should be boolean
                    if obj:
                        def _decorator(call):
                            StartThread(listen, call).start()
                    else:
                        def _decorator(call):
                            self.events[attr] = call
                            return call
                    return _decorator

            return decorator
        elif self.method:
            method = "%s.%s" % (self.method, attr)
            self.method = ""
            return lambda **data: self.call_method(method, data)
        else:
            self.method = attr
            return self



# -*- coding: utf-8 -*-
# author: Ethosa
from json import loads

from websocket import create_connection


class StreamingAPI:
    def __init__(self, vk):
        """Initialize StreamingAPI object.

        Arguments:
            vk {Vk} -- Vk object.
        """
        self.logger = vk.logger
        self.session = vk.session
        self.call_method = vk.call_method

        self.url = ""

    def add_rule(self, tag, value):
        """Adds a new rule in the stream.

        Use this method to add a new rule to the stream.

        Arguments:
            tag {str} -- rule label.
            value {str} -- string representation of the rule.
        """
        obj = {
            "rule": {
                "value": value,
                "tag": tag
            }
        }
        return self.session.post(
            self.url,
            json=obj,
            headers={"Content-Type": "application/json"}
        ).json()

    def auth(self):
        """Logging into the Streaming API.

        Raises:
            ValueError -- wrong access token.

        Returns:
            dict -- result code.
        """
        response = self.call_method("streaming.getServerUrl")
        if "response" in response:
            self.url = "https://%s/rules?key=%s" % (
                response["response"]["endpoint"],
                response["response"]["key"]
            )
            self.endpoint = response["response"]["endpoint"]
            self.key = response["response"]["key"]
        else:
            raise ValueError("%s" % response)

    def delete_rule(self, tag):
        """Removes the rule from the stream.

        Use this method to remove a rule from a stream.

        Arguments:
            tag {str} -- rule label.

        Returns:
            dict -- result code.
        """
        return self.session.delete(
            self.url,
            json={"tag": tag},
            headers={"Content-Type": "application/json"}
        ).json()

    def get_rules(self):
        """Gets the current rules from the stream.

        Use this method to get rules that are already added to the stream.

        Returns:
            dict -- rules.
        """
        return self.session.get(self.url).json()

    def listen(self):
        """Starts websocket listen

        Yields:
            dict -- event
        """
        while 1:
            ws = create_connection(
                "wss://%s/stream?key=%s" % (
                    self.endpoint, self.key
                ),
                header=[
                    "Content-Type: application/json"
                    "Connection: upgrade",
                    "Upgrade: websocket",
                    "Sec-Websocket-Version: 13",
                ]
            )
            result = ws.recv()
            ws.close()
            yield loads(result)



# -*- coding: utf-8 -*-
# author: Ethosa

from json import dumps


class Keyboard(dict):
    def __init__(self, one_time=False, inline=False, other_keyboard=None):
        """Initializes Keyboard object.

        if other_keyboard is another Keyboard object, then params of this keyboard copied from other Keyboard object.

        Keyword Arguments:
            one_time {bool} -- Hides keyboard after the first answer if True. (default: {False})
            inline {bool} -- Shows keyboard inside the message, if True. (default: {False})
            other_keyboard {Keyboard} -- other Keyboard object. (default: {None})
        """
        if other_keyboard:
            self["one_time"] = other_keyboard["one_time"]
            self["inline"] = other_keyboard["inline"]
            self["buttons"] = other_keyboard["buttons"][:]
        else:
            self["inline"] = inline
            self["one_time"] = one_time
            self["buttons"] = [[]]
        if self["inline"]:
            self.max_size = (3, 3)
            del self["one_time"]
        else:
            self.max_size = (4, 10)

    def add(self, button):
        """Adds a new button in keyboard.

        Arguments:
            button {Button} -- Button object.
        """
        if len(self["buttons"]) < self.max_size[1]:
            if len(self["buttons"][-1]) < self.max_size[0]:
                if button["action"]["type"] == "text":
                    self["buttons"][-1].append(button)
                else:
                    if len(self["buttons"][-1]) > 0:
                        self.add_line()
                    self["buttons"][-1].append(button)

    def add_line(self):
        """Adds a new line in the keyboard, if possible."""
        if len(self["buttons"]) < self.max_size[1]:
            self["buttons"].append([])

    def compile(self):
        """Compiles keyboard for sending it in the message.

        Returns:
            dict -- dictionary object for message sending.
        """
        return dumps(self)



# -*- coding: utf-8 -*-
# author: Ethosa
from time import ctime as current_time

from aiohttp import ClientSession, get_event_loop()

from .ALongPoll import ALongPoll
from .AUploader import AUploader
from ..VK.VkAuthManager import VkAuthManager
from ..VK.VkScript import VkScript


class AVk:
    def __init__(self, token="", group_id="",
                 login="", password="", api="5.103",
                 debug=False, loop=get_event_loop()):
        """auth in VK

        Keyword Arguments:
            token {str} -- access_token (default: {""})
            group_id {str} -- group id if you want to log in through the group (default: {""})
            login {str} -- login. used for authorization through the user (default: {""})
            password {str} -- password. used for authorization through the user (default: {""})
            api {str} -- api version (default: {"5.103"})
            debug {bool} -- debug log (default: {False})
            loop {asyncio event loop} -- event loop to use for requests
        """
        self.session = ClientSession(loop=loop)

        # Parses vk.com, if login and password are not empty.
        if login and password:
            self.auth = VkAuthManager()
            self.auth.login(login, password)
            token = self.auth.get_token()

        self.v = api
        self.token = token
        self.group_id = group_id
        self.debug = debug

        self.method = ""
        self.events = {}

        self.longpoll = ALongPoll(self)
        self.uploader = AUploader(self)
        self.vks = VkScript()  # for pyexecute method.

    async def _log(self, logtype, message):
        """
        Outputs log messages.
        """
        if self.debug:
            print("[%s] at %s -- %s" % (
                logtype, current_time(), message)
            )

    async def _wrapper(self, **kwargs):
        """
        Provides convenient usage VK API.
        """
        return await self.call_method(self.method, kwargs)

    async def call_method(self, method, data={}):
        """
        Calls to any method in VK API.

        Arguments:
            method {str} -- method name
            e.g. "messages.send", "wall.post"

        Keyword Arguments:
            data {dict} -- data to send (default: {{}})

        Returns:
            dict -- response after calling method
        """
        self.method = ""
        data["v"] = self.v
        data["access_token"] = self.token
        response = await self.session.post(
                "https://api.vk.com/method/%s" % method, data=data
            )
        response = await response.json()

        # Logging.
        if "error" in response:
            await self._log("ERROR", 'Error [%s] in called method "%s": %s' % (
                    response["error"]["error_code"], method,
                    response["error"]["error_msg"]
                )
            )
        else:
            await self._log("DEBUG", 'Successfully called method "%s"' % (method))
        return response

    async def execute(self, code):
        """
        Calls an execute VK API method

        Arguments:
            code {str} -- VKScript code.

        Returns:
            dict -- response
        """
        return await self.call_method("execute", {"code": code})

    async def pyexecute(self, code):
        """
        Calls an execute VK API method

        Arguments:
            code {str} -- Python code.

        Returns:
            dict -- response
        """
        code = await self.vks.atranslate(code)
        return await self.execute(code)

    async def start_listen(self):
        """
        Starts receiving events from the server.
        """
        async for event in self.longpoll.listen(True):
            if "type" in event:
                if event["type"] in self.events:
                    await self.events[event["type"]](event)
                elif event["type"] in dir(self):
                    await getattr(self, event["type"])(event)
            else:
                self._log("WARNING", "Unknown event passed: %s" % (event))

    def __getattr__(self, attr):
        """
        A convenient alternative for the call_method method.

        Arguments:
            attr {str} -- method name
            e.g. messages.send, wall.post

        Returns:
            response after calling method
        """
        if attr.startswith("on_"):
            attr = attr[3:]

            def _decorator(call):
                self.events[attr] = call
                return call
            return _decorator
        elif self.method:
            self.method = "%s.%s" % (self.method, attr)
            return self._wrapper
        else:
            self.method = attr
            return self



# -*- coding: utf-8 -*-
# author: Ethosa

from .Vk import Vk
from .LongPoll import LongPoll
from .VkAuthManager import VkAuthManager
from .Event import event
from .VkScript import VkScript
from .Uploader import Uploader
from .Button import create_button
from .Keyboard import Keyboard
from .TemplateElement import TemplateElement
from .Template import Template
from .StreamingAPI import StreamingAPI

if __name__ == '__main__':
    print(Vk, LongPoll, VkAuthManager, event, VkScript,
          Uploader, create_button, Keyboard, TemplateElement,
          Template, StreamingAPI)



# -*- coding: utf-8 -*-
# author: Ethosa
from asyncio import sleep

from aiohttp.client_exceptions import ClientOSError

from ..VK.Event import event


class ALongPoll:
    def __init__(self, vk):
        """Class for using longpoll in VK

        Arguments:
            vk {Vk} -- authed VK object
        """
        self.v = vk.v
        self._log = vk._log
        self.debug = vk.debug
        self.session = vk.session
        self.ts = None
        self.server = None
        self.key = None

        self.data = {
            "access_token": vk.token,
            "v": vk.v
        }

        if vk.group_id:
            self.data["group_id"] = vk.group_id
            self.method = "https://api.vk.com/method/groups.getLongPollServer"
            self.for_server = "%s?act=a_check&key=%s&ts=%s&wait=25"
        else:
            self.method = "https://api.vk.com/method/messages.getLongPollServer"
            self.for_server = "https://%s?act=a_check&key=%s&ts=%s&wait=25&mode=202&version=3"

    async def _get_server(self):
        """
        Returns server, ts and key.

        Raises:
            ValueError -- Invalid authentication.
        """
        try:
            response = await self.session.get(self.method, params=self.data)
            response = await response.json()
        except ClientOSError:
            await sleep(.2)
            await self._get_server()
        if "response" in response:
            response = response["response"]
        else:
            raise ValueError("Invalid authentication.")
        self.server = response["server"]
        self.ts = response["ts"]
        self.key = response["key"]

    async def _get_events(self):
        """
        Returns server response after calling.
        """
        try:
            response = await self.session.get(
                self.for_server % (self.server, self.key, self.ts)
            )
            response = await response.json()
            if "ts" not in response or "updates" not in response:
                await self._get_server()
                response = await self._get_events()
            return response
        except ClientOSError:
            await sleep(.2)
            return await self._get_events()

    async def listen(self, ev=False):
        """
        Starts listening.

        Keyword Argments:
            ev {bool} -- always return dict object.

        Yields:
            {dict} -- new event
        """
        # Get server info and check it.
        await self._get_server()

        await self._log("INFO", "LongPoll launched")

        # Start listening.
        while 1:
            response = await self._get_events()
            self.ts = response["ts"]

            for update in response["updates"]:
                if update:
                    if ev:
                        yield event(update)
                    else:
                        yield update



# -*- coding: utf-8 -*-
# author: Ethosa


class TemplateElement(dict):
    def __init__(self, title="", description="",
                 photo_id="", buttons=[], action="open_link",
                 link="https://vk.com"):
        """Initializes a new TemplateElement object.

        Keyword Arguments:
            title {str} (default: {""})
            description {str} (default: {""})
            photo_id {str} -- Id of the picture, which must attach in the template element. (default: {""})
            buttons {list} -- Buttons in TemplateElement. (default: {[]})
            action {str} -- Action of TemplateElement, can be "open_link" or "open_photo". (default: {"open_link"})
            link {str} -- link for "open_link" action. (default: {"https://vk.com"})
        """
        self["title"] = title
        self["description"] = description
        self["photo_id"] = photo_id
        self["buttons"] = buttons
        self["action"] = {"type": action}
        if action == "open_link":
            self["action"]["link"] = link

    def add(self, button):
        """Adds a new Button in TemplateElement.

        Arguments:
            button {Button} -- Button object.
        """
        if len(self["buttons"]) < 3:
            self["buttons"].append(button)



# -*- coding: utf-8 -*-
# author: Ethosa


def create_button(button_type="text", color="primary", **kwargs):
    """
    Creates a new Button object.

    Keyword Arguments:
        button_type {str} -- button type. (default: {"text"})
        color {str} -- button color. (default: {"primary"})
            Can be "primary", "secondary", "positive" or "negative".
    """
    obj = {}
    if button_type == "text":
        obj["color"] = color
    obj["action"] = kwargs
    obj["action"]["type"] = button_type
    return obj



# -*- coding: utf-8 -*-
# author: Ethosa

from .AVk import AVk
from .ALongPoll import ALongPoll
from .AUploader import AUploader
from .AStreamingAPI import AStreamingAPI


if __name__ == '__main__':
    print(AVk, ALongPoll, AUploader, AStreamingAPI)
